import os
from typing import Dict, List

import requests
from fastapi import FastAPI, HTTPException
from fastapi.responses import JSONResponse
from kiln_ai.adapters.ml_model_list import (
    ModelName,
    ModelProviderName,
    built_in_models,
    provider_name_from_id,
    provider_warnings,
)
from kiln_ai.utils.config import Config
from langchain_aws import ChatBedrockConverse
from pydantic import BaseModel


class OllamaConnection(BaseModel):
    message: str
    models: List[str]


async def connect_ollama() -> OllamaConnection:
    # Tags is a list of Ollama models. Proves Ollama is running, and models are available.
    try:
        tags = requests.get("http://localhost:11434/api/tags", timeout=5).json()
    except requests.exceptions.ConnectionError:
        raise HTTPException(
            status_code=417,
            detail="Failed to connect. Ensure Ollama app is running.",
        )
    except Exception as e:
        raise HTTPException(
            status_code=500,
            detail=f"Failed to connect to Ollama: {e}",
        )

    # Build a list of models we support for Ollama from the built-in model list
    supported_ollama_models = [
        provider.provider_options["model"]
        for model in built_in_models
        for provider in model.providers
        if provider.name == ModelProviderName.ollama
    ]

    if "models" in tags:
        models = tags["models"]
        if isinstance(models, list):
            model_names = [model["model"] for model in models]
            available_supported_models = [
                model
                for model in model_names
                if model in supported_ollama_models
                or model in [f"{m}:latest" for m in supported_ollama_models]
            ]
            if available_supported_models:
                Config.shared().ollama_base_url = "http://localhost:11434"
                return OllamaConnection(
                    message="Ollama connected",
                    models=available_supported_models,
                )

    raise HTTPException(
        status_code=417,
        detail="Ollama is running, but no supported models are installed. Install one or more supported model, like 'ollama pull phi3.5'.",
    )


class ModelDetails(BaseModel):
    id: str
    name: str


class AvailableModels(BaseModel):
    provider_name: str
    provider_id: str
    models: List[ModelDetails]


class ProviderModel(BaseModel):
    id: str
    name: str


class ProviderModels(BaseModel):
    models: Dict[ModelName, ProviderModel]


def connect_provider_api(app: FastAPI):
    @app.get("/api/providers/models")
    async def get_providers_models() -> ProviderModels:
        models = {}
        for model in built_in_models:
            models[model.name] = ProviderModel(id=model.name, name=model.friendly_name)
        return ProviderModels(models=models)

    # returns map, of provider name to list of model names
    @app.get("/api/available_models")
    async def get_available_models() -> List[AvailableModels]:
        # Providers with just keys can return all their models if keys are set
        key_providers: List[str] = []

        # Try to connect to Ollama
        try:
            ollama_connection = await connect_ollama()
            if len(ollama_connection.models) > 0:
                key_providers.append(ModelProviderName.ollama)
        except HTTPException:
            # skip ollama if it's not available
            pass

        for provider, provider_warning in provider_warnings.items():
            has_keys = True
            for required_key in provider_warning.required_config_keys:
                if Config.shared().get_value(required_key) is None:
                    has_keys = False
                    break
            if has_keys:
                key_providers.append(provider)
        models: List[AvailableModels] = [
            AvailableModels(
                provider_name=provider_name_from_id(provider),
                provider_id=provider,
                models=[],
            )
            for provider in key_providers
        ]

        for model in built_in_models:
            for provider in model.providers:
                if provider.name in key_providers:
                    available_models = next(
                        (m for m in models if m.provider_id == provider.name), None
                    )
                    if available_models:
                        available_models.models.append(
                            ModelDetails(id=model.name, name=model.friendly_name)
                        )

        return models

    @app.post("/api/provider/ollama/connect")
    async def connect_ollama_api() -> OllamaConnection:
        return await connect_ollama()

    @app.post("/api/provider/connect_api_key")
    async def connect_api_key(payload: dict):
        provider = payload.get("provider")
        key_data = payload.get("key_data")
        if not isinstance(key_data, dict) or not isinstance(provider, str):
            return JSONResponse(
                status_code=400,
                content={"message": "Invalid key_data or provider"},
            )

        api_key_providers = ["openai", "groq", "bedrock", "openrouter"]
        if provider not in api_key_providers:
            return JSONResponse(
                status_code=400,
                content={"message": f"Provider {provider} not supported"},
            )

        if provider == "openai" and isinstance(key_data["API Key"], str):
            return await connect_openai(key_data["API Key"])
        elif provider == "groq" and isinstance(key_data["API Key"], str):
            return await connect_groq(key_data["API Key"])
        elif provider == "openrouter" and isinstance(key_data["API Key"], str):
            return await connect_openrouter(key_data["API Key"])
        elif (
            provider == "bedrock"
            and isinstance(key_data["Access Key"], str)
            and isinstance(key_data["Secret Key"], str)
        ):
            return await connect_bedrock(key_data["Access Key"], key_data["Secret Key"])
        else:
            return JSONResponse(
                status_code=400,
                content={"message": f"Provider {provider} missing API key"},
            )


async def connect_openrouter(key: str):
    try:
        headers = {
            "Authorization": f"Bearer {key}",
            "Content-Type": "application/json",
        }
        # invalid body, but we just want to see if the key is valid
        response = requests.post(
            "https://openrouter.ai/api/v1/chat/completions",
            headers=headers,
            json={},
        )

        # 401 def means invalid API key
        if response.status_code == 401:
            return JSONResponse(
                status_code=401,
                content={
                    "message": "Failed to connect to OpenRouter. Invalid API key."
                },
            )
        else:
            # No 401 means key is valid (even it it's an error, which we expect with empty body)
            Config.shared().open_router_api_key = key

            return JSONResponse(
                status_code=200,
                content={"message": "Connected to OpenRouter"},
            )
            # Any non-200 status code is an error
    except Exception as e:
        # unexpected error
        return JSONResponse(
            status_code=400,
            content={"message": f"Failed to connect to OpenRouter. Error: {str(e)}"},
        )


async def connect_openai(key: str):
    try:
        headers = {
            "Authorization": f"Bearer {key}",
            "Content-Type": "application/json",
        }
        response = requests.get("https://api.openai.com/v1/models", headers=headers)

        # 401 def means invalid API key, so special case it
        if response.status_code == 401:
            return JSONResponse(
                status_code=401,
                content={"message": "Failed to connect to OpenAI. Invalid API key."},
            )

        # Any non-200 status code is an error
        response.raise_for_status()
        # If the request is successful, the function will continue
    except Exception as e:
        return JSONResponse(
            status_code=400,
            content={"message": f"Failed to connect to OpenAI. Error: {str(e)}"},
        )

    # It worked! Save the key and return success
    Config.shared().open_ai_api_key = key

    return JSONResponse(
        status_code=200,
        content={"message": "Connected to OpenAI"},
    )


async def connect_groq(key: str):
    try:
        headers = {
            "Authorization": f"Bearer {key}",
            "Content-Type": "application/json",
        }
        response = requests.get(
            "https://api.groq.com/openai/v1/models", headers=headers
        )

        if "invalid_api_key" in response.text:
            return JSONResponse(
                status_code=401,
                content={"message": "Failed to connect to Groq. Invalid API key."},
            )

        # Any non-200 status code is an error
        response.raise_for_status()
        # If the request is successful, the function will continue
    except Exception as e:
        return JSONResponse(
            status_code=400,
            content={"message": f"Failed to connect to Groq. Error: {str(e)}"},
        )

    # It worked! Save the key and return success
    Config.shared().groq_api_key = key

    return JSONResponse(
        status_code=200,
        content={"message": "Connected to Groq"},
    )


async def connect_bedrock(access_key: str, secret_key: str):
    try:
        # Langchain API is not good... need to use env vars. Pop these in finally block
        os.environ["AWS_ACCESS_KEY_ID"] = access_key
        os.environ["AWS_SECRET_ACCESS_KEY"] = secret_key

        # Fake model, but will get a credential error before AWS realizes it's wrong
        # Ensures we don't spend money on a test call
        llm = ChatBedrockConverse(
            model="fake_model",
            region_name="us-west-2",
        )
        llm.invoke("Hello, how are you?")
    except Exception as e:
        # Check for specific error messages indicating invalid credentials
        if "UnrecognizedClientException" in str(
            e
        ) or "InvalidSignatureException" in str(e):
            return JSONResponse(
                status_code=401,
                content={
                    "message": "Failed to connect to Bedrock. Invalid credentials."
                },
            )
        else:
            # We still expect an error (fake model), but not for invalid credentials which means success!
            Config.shared().bedrock_access_key = access_key
            Config.shared().bedrock_secret_key = secret_key
            return JSONResponse(
                status_code=200,
                content={"message": "Connected to Bedrock"},
            )

    finally:
        os.environ.pop("AWS_ACCESS_KEY_ID")
        os.environ.pop("AWS_SECRET_ACCESS_KEY")

    # we shouldn't get here, but if we do, something went wrong
    return JSONResponse(
        status_code=400,
        content={"message": "Unknown Bedrock Error"},
    )



from fastapi import FastAPI
from kiln_ai.utils.config import Config


def connect_settings(app: FastAPI):
    @app.post("/api/settings")
    def update_settings(new_settings: dict[str, int | float | str | bool | None]):
        Config.shared().update_settings(new_settings)
        return Config.shared().settings(hide_sensitive=True)

    @app.get("/api/settings")
    def read_settings():
        settings = Config.shared().settings(hide_sensitive=True)
        return settings

    @app.get("/api/settings/{item_id}")
    def read_item(item_id: str):
        settings = Config.shared().settings(hide_sensitive=True)
        return {item_id: settings.get(item_id, None)}



import os
from unittest.mock import patch

import pytest
import yaml
from fastapi import FastAPI
from fastapi.testclient import TestClient
from kiln_ai.utils.config import Config

from libs.server.kiln_server.settings import connect_settings


@pytest.fixture
def temp_home(tmp_path):
    with patch.object(os.path, "expanduser", return_value=str(tmp_path)):
        yield tmp_path


@pytest.fixture
def app():
    app = FastAPI()
    connect_settings(app)
    return app


@pytest.fixture
def client(app):
    return TestClient(app)


def test_load_settings_empty(temp_home):
    assert Config.shared().load_settings() == {}


def test_load_settings_existing(temp_home):
    settings_file = Config.settings_path()
    os.makedirs(os.path.dirname(settings_file), exist_ok=True)
    test_settings = {"key": "value"}
    with open(settings_file, "w") as f:
        yaml.dump(test_settings, f)

    assert Config.shared().load_settings() == test_settings


def test_update_settings(client, temp_home):
    new_settings = {"test_key": "test_value", "test_key2": "test_value2"}
    response = client.post("/api/settings", json=new_settings)
    assert response.status_code == 200
    assert response.json() == new_settings

    # Verify the settings were actually updated
    with open(Config.settings_path(), "r") as f:
        saved_settings = yaml.safe_load(f)
    assert saved_settings == new_settings


def test_read_settings(client, temp_home):
    test_settings = {"key1": "value1", "key2": 42}
    Config.shared().update_settings(test_settings)

    response = client.get("/api/settings")
    assert response.status_code == 200
    assert response.json() == test_settings


def test_read_item(client, temp_home):
    response = client.get("/api/settings/setting1")
    assert response.status_code == 200
    assert response.json() == {"setting1": None}
    Config.shared().update_settings({"setting1": "value1"})
    response = client.get("/api/settings/setting1")
    assert response.status_code == 200
    assert response.json() == {"setting1": "value1"}


def test_clear_setting(client, temp_home):
    # First, set a value
    initial_settings = {"test_key": "test_value"}
    Config.shared().update_settings(initial_settings)

    # Verify the setting was set
    response = client.get("/api/settings/test_key")
    assert response.status_code == 200
    assert response.json() == {"test_key": "test_value"}

    # Now, clear the setting by posting a null value
    clear_settings = {"test_key": None}
    response = client.post("/api/settings", json=clear_settings)
    assert response.status_code == 200
    assert response.json() == {}

    # Verify the setting was cleared
    response = client.get("/api/settings/test_key")
    assert response.status_code == 200
    assert response.json() == {"test_key": None}

    # Check the full settings to ensure the key was removed
    response = client.get("/api/settings")
    assert response.status_code == 200
    assert "test_key" not in response.json()


@pytest.fixture
def mock_config(monkeypatch):
    mock_settings = {
        "public_setting": "visible",
        "sensitive_setting": "secret",
    }

    class MockConfig:
        @staticmethod
        def shared():
            return MockConfig()

        def settings(self, hide_sensitive=False):
            if hide_sensitive:
                return {
                    k: "[hidden]" if k == "sensitive_setting" else v
                    for k, v in mock_settings.items()
                }
            return mock_settings

        def update_settings(self, new_settings):
            mock_settings.update(new_settings)

    monkeypatch.setattr(Config, "shared", MockConfig.shared)
    return MockConfig()


# Confirm secrets are hidden
def test_settings_endpoints(client, mock_config):
    # Test GET /settings
    response = client.get("/api/settings")
    assert response.status_code == 200
    assert response.json() == {
        "public_setting": "visible",
        "sensitive_setting": "[hidden]",
    }

    # Test POST /settings
    new_settings = {"public_setting": "new_value", "sensitive_setting": "new_secret"}
    response = client.post("/api/settings", json=new_settings)
    assert response.status_code == 200
    assert response.json() == {
        "public_setting": "new_value",
        "sensitive_setting": "[hidden]",
    }

    # Test GET /settings/{item_id}
    response = client.get("/api/settings/public_setting")
    assert response.status_code == 200
    assert response.json() == {"public_setting": "new_value"}

    response = client.get("/api/settings/sensitive_setting")
    assert response.status_code == 200
    assert response.json() == {"sensitive_setting": "[hidden]"}



import json
from unittest.mock import AsyncMock, patch

import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient

from libs.core.kiln_ai.datamodel import (
    DataSource,
    DataSourceType,
    Project,
    Task,
    TaskOutput,
    TaskRun,
)
from libs.server.kiln_server.repair_api import (
    RepairRunPost,
    RepairTaskApiInput,
    connect_repair_api,
)


@pytest.fixture
def app():
    app = FastAPI()
    connect_repair_api(app)
    return app


@pytest.fixture
def client(app):
    return TestClient(app)


@pytest.fixture
def data_source():
    return DataSource(
        type=DataSourceType.synthetic,
        properties={
            "model_name": "gpt_4o",
            "model_provider": "openai",
            "adapter_name": "langchain_adapter",
            "prompt_builder_name": "simple_prompt_builder",
        },
    )


@pytest.fixture
def improvement_task(tmp_path) -> Task:
    project_path = tmp_path / "test_project" / "project.kiln"
    project_path.parent.mkdir()

    project = Project(name="Test Project", path=str(project_path))
    project.save_to_file()

    task = Task(name="Test Task", instruction="Test Instruction", parent=project)
    task.save_to_file()
    return task


@pytest.fixture
def improvement_task_run(data_source, improvement_task):
    task_run = TaskRun(
        input="Test Input",
        input_source=data_source,
        parent=improvement_task,
        output=TaskOutput(output="Test Output", source=data_source),
    )
    task_run.save_to_file()

    return task_run


@pytest.fixture
def mock_repair_task_run(improvement_task, data_source):
    return TaskRun(
        output=TaskOutput(
            output="Test Output",
            source=data_source,
        ),
        input="The input to the improvement task",
        input_source=data_source,
    )


@pytest.fixture
def mock_langchain_adapter(mock_repair_task_run):
    with patch("libs.server.kiln_server.repair_api.LangChainPromptAdapter") as mock:
        mock_adapter = AsyncMock()
        mock_adapter.invoke = AsyncMock()
        mock.return_value = mock_adapter

        mock_adapter.invoke.return_value = mock_repair_task_run

        yield mock_adapter


@pytest.fixture
def mock_run_and_task(improvement_task, improvement_task_run):
    with patch("libs.server.kiln_server.repair_api.task_and_run_from_id") as mock:
        mock.return_value = (improvement_task, improvement_task_run)
        yield mock


def test_repair_run_success(
    mock_run_and_task,
    mock_langchain_adapter,
    client,
    improvement_task,
    mock_repair_task_run,
):
    # Arrange
    input_data = RepairTaskApiInput(evaluator_feedback="Fix this issue")

    # Act
    response = client.post(
        "/api/projects/proj-ID/tasks/task-ID/runs/run-ID/run_repair",
        json=input_data.model_dump(),
    )

    # Assert
    assert response.status_code == 200
    res = response.json()
    repaired_output = res["output"]
    assert TaskOutput.model_validate(repaired_output) == mock_repair_task_run.output
    mock_langchain_adapter.invoke.assert_awaited_once()


def test_save_repair_success(
    mock_run_and_task,
    mock_langchain_adapter,
    client,
    improvement_task,
    mock_repair_task_run,
):
    # Arrange
    input_data = RepairRunPost(
        repair_run=mock_repair_task_run.model_dump(),
        evaluator_feedback="Fix this issue",
    )

    # Act
    response = client.post(
        "/api/projects/proj-ID/tasks/task-ID/runs/run-ID/repair",
        json=json.loads(input_data.model_dump_json()),
    )

    # Assert
    res = response.json()
    assert response.status_code == 200
    repaired_output = res["repaired_output"]
    assert TaskOutput.model_validate(repaired_output) == mock_repair_task_run.output
    assert res["repair_instructions"] == "Fix this issue"
    assert res["input"] == "Test Input"

    # Verify that the run was updated in the file system
    updated_run = improvement_task.runs()[0]
    assert updated_run.repair_instructions == "Fix this issue"
    assert updated_run.repaired_output == mock_repair_task_run.output


def test_repair_run_missing_model_info(
    mock_run_and_task,
    mock_langchain_adapter,
    client,
    improvement_task,
    mock_repair_task_run,
):
    # Arrange
    input_data = RepairTaskApiInput(evaluator_feedback="Fix this issue")
    mock_run_and_task.return_value[1].output.source.properties["model_name"] = None
    mock_run_and_task.return_value[1].output.source.properties["model_provider"] = None

    # Act
    response = client.post(
        "/api/projects/proj-ID/tasks/task-ID/runs/run-ID/run_repair",
        json=input_data.model_dump(),
    )

    # Assert
    assert response.status_code == 400
    assert response.json()["detail"] == "Model name and provider must be specified."



from unittest.mock import MagicMock, patch

import pytest

# Create a FastAPI app and connect the prompt_api
from fastapi import FastAPI
from fastapi.testclient import TestClient
from kiln_ai.datamodel import Task

from libs.core.kiln_ai.adapters.prompt_builders import BasePromptBuilder
from libs.server.kiln_server.prompt_api import connect_prompt_api

app = FastAPI()
connect_prompt_api(app)

# Create a test client
client = TestClient(app)


# Mock prompt builder class
class MockPromptBuilder(BasePromptBuilder):
    @classmethod
    def prompt_builder_name(cls):
        return "MockPromptBuilder"

    def build_prompt(self):
        return "Mock prompt"


@pytest.fixture
def mock_task():
    return MagicMock(spec=Task)


@pytest.fixture
def mock_prompt_builder_from_ui_name():
    with patch(
        "libs.server.kiln_server.prompt_api.prompt_builder_from_ui_name"
    ) as mock:
        mock.return_value = MockPromptBuilder
        yield mock


@pytest.fixture
def mock_task_from_id():
    with patch("libs.server.kiln_server.prompt_api.task_from_id") as mock:
        mock.return_value = MagicMock(spec=Task)
        yield mock


def test_gen_prompt_success(
    mock_task, mock_prompt_builder_from_ui_name, mock_task_from_id
):
    response = client.get(
        "/api/projects/project123/task/task456/gen_prompt/mock_generator"
    )

    assert response.status_code == 200
    data = response.json()
    assert data == {
        "prompt": "Mock prompt",
        "prompt_builder_name": "MockPromptBuilder",
        "ui_generator_name": "mock_generator",
    }

    mock_task_from_id.assert_called_once_with("project123", "task456")
    mock_prompt_builder_from_ui_name.assert_called_once_with("mock_generator")


def test_gen_prompt_exception(
    mock_task, mock_prompt_builder_from_ui_name, mock_task_from_id
):
    mock_prompt_builder_from_ui_name.side_effect = ValueError(
        "Invalid prompt generator"
    )

    response = client.get(
        "/api/projects/project123/task/task456/gen_prompt/invalid_generator"
    )

    assert response.status_code == 400
    data = response.json()
    assert data == {"detail": "Invalid prompt generator"}

    mock_task_from_id.assert_called_once_with("project123", "task456")
    mock_prompt_builder_from_ui_name.assert_called_once_with("invalid_generator")



import os
import tempfile
from unittest.mock import MagicMock, patch

import pytest
import requests
from fastapi import HTTPException
from fastapi.testclient import TestClient

from libs.server.kiln_server.server import make_app
from libs.server.kiln_server.webhost import HTMLStaticFiles


@pytest.fixture
def client():
    # a client based on a mock studio path
    with tempfile.TemporaryDirectory() as temp_dir:
        os.makedirs(temp_dir, exist_ok=True)
        with patch("libs.server.kiln_server.webhost.studio_path", new=lambda: temp_dir):
            from libs.server.kiln_server.webhost import studio_path

            assert studio_path() == temp_dir  # Verify the patch is working
            app = make_app()
            client = TestClient(app)
            yield client


def test_ping(client):
    response = client.get("/ping")
    assert response.status_code == 200
    assert response.json() == "pong"


def test_connect_ollama_success(client):
    with patch("requests.get") as mock_get:
        mock_get.return_value.json.return_value = {
            "models": [{"model": "phi3.5:latest"}]
        }
        response = client.post("/api/provider/ollama/connect")
        assert response.status_code == 200
        assert response.json() == {
            "message": "Ollama connected",
            "models": ["phi3.5:latest"],
        }


def test_connect_ollama_connection_error(client):
    with patch("requests.get") as mock_get:
        mock_get.side_effect = requests.exceptions.ConnectionError
        response = client.post("/api/provider/ollama/connect")
        assert response.status_code == 417
        assert response.json() == {
            "message": "Failed to connect. Ensure Ollama app is running."
        }


def test_connect_ollama_general_exception(client):
    with patch("requests.get") as mock_get:
        mock_get.side_effect = Exception("Test exception")
        response = client.post("/api/provider/ollama/connect")
        assert response.status_code == 500
        assert response.json() == {
            "message": "Failed to connect to Ollama: Test exception"
        }


def test_connect_ollama_no_models(client):
    with patch("requests.get") as mock_get:
        mock_get.return_value.json.return_value = {"models": []}
        response = client.post("/api/provider/ollama/connect")
        assert response.status_code == 417
        assert (
            "Ollama is running, but no supported models are installed"
            in response.json()["message"]
        )


@pytest.mark.parametrize(
    "origin",
    [
        "http://localhost",
        "http://localhost:8000",
        "http://127.0.0.1",
        "http://127.0.0.1:3000",
        "https://localhost:8443",
        "https://127.0.0.1:8443",
    ],
)
def test_cors_allowed_origins(client, origin):
    response = client.get("/ping", headers={"Origin": origin})
    assert response.status_code == 200
    assert response.headers["access-control-allow-origin"] == origin


@pytest.mark.parametrize(
    "origin",
    [
        "http://example.com",
        "https://kiln-ai.com",
        "http://192.168.1.100",
        "http://localhost.com",
        "http://127.0.0.2",
        "http://127.0.0.2.com",
    ],
)
def test_cors_blocked_origins(client, origin):
    response = client.get("/ping", headers={"Origin": origin})
    assert response.status_code == 200
    assert "access-control-allow-origin" not in response.headers


def create_studio_test_file(relative_path):
    from libs.server.kiln_server.webhost import studio_path

    full_path = os.path.join(studio_path(), relative_path)
    os.makedirs(os.path.dirname(full_path), exist_ok=True)
    with open(full_path, "w") as f:
        f.write("<html><body>Test</body></html>")
    return full_path


def test_cors_no_origin(client):
    # Create index.html in the mock studio path
    create_studio_test_file("index.html")

    # Use the client to get the root path
    response = client.get("/")

    # Assert the response
    assert response.status_code == 200
    assert "access-control-allow-origin" not in response.headers


class TestHTMLStaticFiles:
    @pytest.fixture
    def html_static_files(self):
        import os
        import tempfile

        self.test_dir = tempfile.mkdtemp()
        with open(os.path.join(self.test_dir, "existing_file"), "w") as f:
            f.write("Test content")
        return HTMLStaticFiles(directory=self.test_dir, html=True)

    @pytest.mark.asyncio
    async def test_get_response_existing_file(self, html_static_files):
        with patch("fastapi.staticfiles.StaticFiles.get_response") as mock_get_response:
            mock_response = MagicMock()
            mock_get_response.return_value = mock_response

            response = await html_static_files.get_response("existing_file", {})

            assert response == mock_response
            mock_get_response.assert_called_once_with("existing_file", {})

    @pytest.mark.asyncio
    async def test_get_response_html_fallback(self, html_static_files):
        with patch("fastapi.staticfiles.StaticFiles.get_response") as mock_get_response:

            def side_effect(path, scope):
                if path.endswith(".html"):
                    return MagicMock()
                raise HTTPException(status_code=404)

            mock_get_response.side_effect = side_effect

            response = await html_static_files.get_response("non_existing_file", {})

            assert response is not None
            assert mock_get_response.call_count == 2
            mock_get_response.assert_any_call("non_existing_file", {})
            mock_get_response.assert_any_call("non_existing_file.html", {})

    @pytest.mark.asyncio
    async def test_get_response_not_found(self, html_static_files):
        with patch("fastapi.staticfiles.StaticFiles.get_response") as mock_get_response:
            mock_get_response.side_effect = HTTPException(status_code=404)

            with pytest.raises(HTTPException):
                await html_static_files.get_response("non_existing_file", {})


@pytest.mark.asyncio
async def test_setup_route(client):
    # Ensure studio_path exists
    create_studio_test_file("index.html")
    create_studio_test_file("path.html")
    create_studio_test_file("nested/index.html")

    # root index.html
    response = client.get("/")
    assert response.status_code == 200
    assert response.text == "<html><body>Test</body></html>"
    # setup.html
    response = client.get("/path")
    assert response.status_code == 200
    assert response.text == "<html><body>Test</body></html>"
    # nested index.html
    response = client.get("/nested")
    assert response.status_code == 200
    assert response.text == "<html><body>Test</body></html>"
    # non existing file

    # expected 404
    with pytest.raises(Exception):
        client.get("/non_existing_file")
    with pytest.raises(Exception):
        client.get("/nested/non_existing_file")



from typing import Any, Dict, List

from fastapi import FastAPI, HTTPException
from kiln_ai.datamodel import Task
from pydantic import BaseModel

from libs.core.kiln_ai.adapters.prompt_builders import prompt_builder_from_ui_name
from libs.server.kiln_server.project_api import project_from_id
from libs.server.kiln_server.task_api import task_from_id


class PromptApiResponse(BaseModel):
    prompt: str
    prompt_builder_name: str
    ui_generator_name: str


def connect_prompt_api(app: FastAPI):
    @app.get("/api/projects/{project_id}/task/{task_id}/gen_prompt/{prompt_generator}")
    async def gen_prompt(
        project_id: str, task_id: str, prompt_generator: str
    ) -> PromptApiResponse:
        task = task_from_id(project_id, task_id)

        try:
            prompt_builder_class = prompt_builder_from_ui_name(prompt_generator)
            prompt_builder = prompt_builder_class(task)
            prompt = prompt_builder.build_prompt()
        except Exception as e:
            raise HTTPException(status_code=400, detail=str(e))

        return PromptApiResponse(
            prompt=prompt,
            prompt_builder_name=prompt_builder_class.prompt_builder_name(),
            ui_generator_name=prompt_generator,
        )



from fastapi import FastAPI, HTTPException
from kiln_ai.datamodel import TaskRun
from pydantic import BaseModel, ConfigDict, Field

from libs.core.kiln_ai.adapters.langchain_adapters import LangChainPromptAdapter
from libs.core.kiln_ai.adapters.repair.repair_task import RepairTaskRun
from libs.server.kiln_server.run_api import task_and_run_from_id


class RepairTaskApiInput(BaseModel):
    evaluator_feedback: str = Field(
        description="Feedback from an evaluator on how to repair the task run."
    )
    model_name: str | None = Field(
        description="The name of the model to use for the repair task. Optional, if not specified, the model of the original task will be used.",
        default=None,
    )
    provider: str | None = Field(
        description="The provider of the model to use for the repair task. Optional, if not specified, the provider of the original task will be used.",
        default=None,
    )

    # Allows use of the model_name field (usually pydantic will reserve model_*)
    model_config = ConfigDict(protected_namespaces=())


class RepairRunPost(BaseModel):
    repair_run: TaskRun
    evaluator_feedback: str


def connect_repair_api(app: FastAPI):
    @app.post("/api/projects/{project_id}/tasks/{task_id}/runs/{run_id}/run_repair")
    async def run_repair(
        project_id: str, task_id: str, run_id: str, input: RepairTaskApiInput
    ) -> TaskRun:
        task, run = task_and_run_from_id(project_id, task_id, run_id)
        repair_task = RepairTaskRun(task)
        repair_task_input = RepairTaskRun.build_repair_task_input(
            original_task=task,
            task_run=run,
            evaluator_feedback=input.evaluator_feedback,
        )

        model_name = input.model_name or run.output.source.properties.get("model_name")
        provider = input.provider or run.output.source.properties.get("model_provider")
        if (
            not model_name
            or not provider
            or not isinstance(model_name, str)
            or not isinstance(provider, str)
        ):
            raise HTTPException(
                status_code=400,
                detail="Model name and provider must be specified.",
            )

        adapter = LangChainPromptAdapter(
            repair_task, model_name=model_name, provider=provider
        )

        repair_run = await adapter.invoke(repair_task_input.model_dump())
        return repair_run

    @app.post("/api/projects/{project_id}/tasks/{task_id}/runs/{run_id}/repair")
    async def post_repair_run(
        project_id: str, task_id: str, run_id: str, input: RepairRunPost
    ) -> TaskRun:
        task, run = task_and_run_from_id(project_id, task_id, run_id)
        # Update the run object atomically, as validation will fail setting one at a time.
        updated_data = run.model_dump()
        updated_data.update(
            {
                "repair_instructions": input.evaluator_feedback,
                "repaired_output": input.repair_run.output,
            }
        )
        updated_run = TaskRun.model_validate(updated_data)
        updated_run.path = run.path

        # Save the updated run
        updated_run.save_to_file()
        return updated_run



import os
import sys

from fastapi import FastAPI
from fastapi.responses import FileResponse
from fastapi.staticfiles import StaticFiles


def studio_path():
    try:
        # pyinstaller path
        base_path = sys._MEIPASS  # type: ignore
        return os.path.join(base_path, "./web_ui/build")
    except Exception:
        base_path = os.path.join(os.path.dirname(__file__), "..")
        return os.path.join(base_path, "../../app/web_ui/build")


# File server that maps /foo/bar to /foo/bar.html (Starlette StaticFiles only does index.html)
class HTMLStaticFiles(StaticFiles):
    async def get_response(self, path: str, scope):
        try:
            response = await super().get_response(path, scope)
            if response.status_code != 404:
                return response
        except Exception as e:
            # catching HTTPException explicitly not working for some reason
            if getattr(e, "status_code", None) != 404:
                # Don't raise on 404, fall through to return the .html version
                raise e
        #  Try the .html version of the file if the .html version exists, for 404s
        return await super().get_response(f"{path}.html", scope)


def connect_webhost(app: FastAPI):
    # Ensure studio_path exists (test servers don't necessarily create it)
    os.makedirs(studio_path(), exist_ok=True)
    # Serves the web UI at root
    app.mount("/", HTMLStaticFiles(directory=studio_path(), html=True), name="studio")

    # add pretty 404s
    @app.exception_handler(404)
    def not_found_exception_handler(request, exc):
        # don't handle /api routes, which return JSON errors
        if request.url.path.startswith("/api"):
            raise exc
        return FileResponse(os.path.join(studio_path(), "404.html"), status_code=404)



import json
import os
from unittest.mock import MagicMock, patch

import pytest
from fastapi import FastAPI, HTTPException
from fastapi.responses import JSONResponse
from fastapi.testclient import TestClient
from kiln_ai.adapters.ml_model_list import (
    KilnModel,
    KilnModelProvider,
    ModelName,
    ModelProviderName,
    built_in_models,
)
from kiln_ai.utils.config import Config

from libs.server.kiln_server.provider_api import (
    ChatBedrockConverse,
    OllamaConnection,
    connect_bedrock,
    connect_groq,
    connect_openrouter,
    connect_provider_api,
)


@pytest.fixture
def app():
    app = FastAPI()
    connect_provider_api(app)
    return app


@pytest.fixture
def client(app):
    return TestClient(app)


def test_connect_api_key_invalid_payload(client):
    response = client.post(
        "/api/provider/connect_api_key",
        json={"provider": "openai", "key_data": "invalid"},
    )
    assert response.status_code == 400
    assert response.json() == {"message": "Invalid key_data or provider"}


def test_connect_api_key_unsupported_provider(client):
    response = client.post(
        "/api/provider/connect_api_key",
        json={"provider": "unsupported", "key_data": {"API Key": "test"}},
    )
    assert response.status_code == 400
    assert response.json() == {"message": "Provider unsupported not supported"}


@patch("libs.server.kiln_server.provider_api.connect_openai")
def test_connect_api_key_openai_success(mock_connect_openai, client):
    mock_connect_openai.return_value = {"message": "Connected to OpenAI"}
    response = client.post(
        "/api/provider/connect_api_key",
        json={"provider": "openai", "key_data": {"API Key": "test_key"}},
    )
    assert response.status_code == 200
    assert response.json() == {"message": "Connected to OpenAI"}
    mock_connect_openai.assert_called_once_with("test_key")


@patch("libs.server.kiln_server.provider_api.requests.get")
@patch("libs.server.kiln_server.provider_api.Config.shared")
def test_connect_openai_success(mock_config_shared, mock_requests_get, client):
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_requests_get.return_value = mock_response

    mock_config = MagicMock()
    mock_config_shared.return_value = mock_config

    response = client.post(
        "/api/provider/connect_api_key",
        json={"provider": "openai", "key_data": {"API Key": "test_key"}},
    )

    assert response.status_code == 200
    assert response.json() == {"message": "Connected to OpenAI"}
    assert mock_config.open_ai_api_key == "test_key"


@patch("libs.server.kiln_server.provider_api.requests.get")
def test_connect_openai_invalid_key(mock_requests_get, client):
    mock_response = MagicMock()
    mock_response.status_code = 401
    mock_requests_get.return_value = mock_response

    response = client.post(
        "/api/provider/connect_api_key",
        json={"provider": "openai", "key_data": {"API Key": "invalid_key"}},
    )

    assert response.status_code == 401
    assert response.json() == {
        "message": "Failed to connect to OpenAI. Invalid API key."
    }


@patch("libs.server.kiln_server.provider_api.requests.get")
def test_connect_openai_request_exception(mock_requests_get, client):
    mock_requests_get.side_effect = Exception("Test error")

    response = client.post(
        "/api/provider/connect_api_key",
        json={"provider": "openai", "key_data": {"API Key": "test_key"}},
    )

    assert response.status_code == 400
    assert "Failed to connect to OpenAI. Error:" in response.json()["message"]


@pytest.fixture
def mock_requests_get():
    with patch("libs.server.kiln_server.provider_api.requests.get") as mock_get:
        yield mock_get


@pytest.fixture
def mock_config():
    with patch("libs.server.kiln_server.provider_api.Config") as mock_config:
        mock_config.shared.return_value = MagicMock()
        yield mock_config


@patch("libs.server.kiln_server.provider_api.requests.get")
@patch("libs.server.kiln_server.provider_api.Config.shared")
async def test_connect_groq_success(mock_config_shared, mock_requests_get):
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.text = '{"models": []}'
    mock_requests_get.return_value = mock_response

    mock_config = MagicMock()
    mock_config_shared.return_value = mock_config

    assert mock_config.shared.return_value.groq_api_key != "test_api_key"
    result = await connect_groq("test_api_key")

    assert result.status_code == 200
    assert result.body == b'{"message":"Connected to Groq"}'
    mock_config.shared.return_value.groq_api_key = "test_api_key"
    mock_requests_get.assert_called_once_with(
        "https://api.groq.com/openai/v1/models",
        headers={
            "Authorization": "Bearer test_api_key",
            "Content-Type": "application/json",
        },
    )
    assert mock_config.shared.return_value.groq_api_key == "test_api_key"


async def test_connect_groq_invalid_api_key(mock_requests_get):
    mock_response = MagicMock()
    mock_response.status_code = 401
    mock_response.text = "{a:'invalid_api_key'}"
    mock_requests_get.return_value = mock_response

    result = await connect_groq("invalid_key")

    assert result.status_code == 401
    response_data = json.loads(result.body)
    assert "Invalid API key" in response_data["message"]


async def test_connect_groq_request_error(mock_requests_get):
    mock_requests_get.side_effect = Exception("Connection error")

    result = await connect_groq("test_api_key")

    assert result.status_code == 400
    response_data = json.loads(result.body)
    assert "Failed to connect to Groq" in response_data["message"]


async def test_connect_groq_non_200_response(mock_requests_get):
    mock_response = MagicMock()
    mock_response.status_code = 500
    mock_response.raise_for_status.side_effect = Exception("Server error")
    mock_requests_get.return_value = mock_response

    result = await connect_groq("test_api_key")

    assert result.status_code == 400
    response_data = json.loads(result.body)
    assert "Failed to connect to Groq" in response_data["message"]


@pytest.mark.asyncio
async def test_connect_openrouter():
    # Test case 1: Valid API key
    with patch("requests.post") as mock_post:
        mock_response = MagicMock()
        mock_response.status_code = (
            400  # Simulating an expected error due to empty body
        )
        mock_post.return_value = mock_response

        result = await connect_openrouter("valid_api_key")
        assert result.status_code == 200
        assert result.body == b'{"message":"Connected to OpenRouter"}'
        assert Config.shared().open_router_api_key == "valid_api_key"

    # Test case 2: Invalid API key
    with patch("requests.post") as mock_post:
        mock_response = MagicMock()
        mock_response.status_code = 401
        mock_post.return_value = mock_response

        result = await connect_openrouter("invalid_api_key")
        assert result.status_code == 401
        assert (
            result.body
            == b'{"message":"Failed to connect to OpenRouter. Invalid API key."}'
        )
        assert Config.shared().open_router_api_key != "invalid_api_key"

    # Test case 3: Unexpected error
    with patch("requests.post") as mock_post:
        mock_post.side_effect = Exception("Unexpected error")

        result = await connect_openrouter("api_key")
        assert result.status_code == 400
        assert (
            b"Failed to connect to OpenRouter. Error: Unexpected error" in result.body
        )
        assert Config.shared().open_router_api_key != "api_key"


@pytest.fixture
def mock_environ():
    with patch("libs.server.kiln_server.provider_api.os.environ", {}) as mock_env:
        yield mock_env


@pytest.mark.asyncio
@patch("libs.server.kiln_server.provider_api.ChatBedrockConverse")
async def test_connect_bedrock_success(mock_chat_bedrock, mock_environ):
    mock_llm = MagicMock()
    mock_chat_bedrock.return_value = mock_llm
    mock_llm.invoke.side_effect = Exception("Some non-credential error")

    result = await connect_bedrock("test_access_key", "test_secret_key")

    assert isinstance(result, JSONResponse)
    assert result.status_code == 200
    assert result.body == b'{"message":"Connected to Bedrock"}'
    assert Config.shared().bedrock_access_key == "test_access_key"
    assert Config.shared().bedrock_secret_key == "test_secret_key"

    mock_chat_bedrock.assert_called_once_with(
        model="fake_model",
        region_name="us-west-2",
    )
    mock_llm.invoke.assert_called_once_with("Hello, how are you?")


@pytest.mark.asyncio
@patch("libs.server.kiln_server.provider_api.ChatBedrockConverse")
async def test_connect_bedrock_invalid_credentials(mock_chat_bedrock, mock_environ):
    mock_llm = MagicMock()
    mock_chat_bedrock.return_value = mock_llm
    mock_llm.invoke.side_effect = Exception("UnrecognizedClientException")

    result = await connect_bedrock("invalid_access_key", "invalid_secret_key")

    assert isinstance(result, JSONResponse)
    assert result.status_code == 401
    assert (
        result.body
        == b'{"message":"Failed to connect to Bedrock. Invalid credentials."}'
    )

    assert "AWS_ACCESS_KEY_ID" not in mock_environ
    assert "AWS_SECRET_ACCESS_KEY" not in mock_environ


@pytest.mark.asyncio
@patch("libs.server.kiln_server.provider_api.ChatBedrockConverse")
async def test_connect_bedrock_unknown_error(mock_chat_bedrock, mock_environ):
    mock_llm = MagicMock()
    mock_chat_bedrock.return_value = mock_llm
    mock_llm.invoke.side_effect = Exception("Some unexpected error")

    result = await connect_bedrock("test_access_key", "test_secret_key")

    assert isinstance(result, JSONResponse)
    assert result.status_code == 200
    assert result.body == b'{"message":"Connected to Bedrock"}'
    assert Config.shared().bedrock_access_key == "test_access_key"
    assert Config.shared().bedrock_secret_key == "test_secret_key"


@pytest.mark.asyncio
@patch("libs.server.kiln_server.provider_api.ChatBedrockConverse")
async def test_connect_bedrock_environment_variables(mock_chat_bedrock, mock_environ):
    mock_llm = MagicMock()
    mock_chat_bedrock.return_value = mock_llm
    mock_llm.invoke.side_effect = Exception("Some non-credential error")

    await connect_bedrock("test_access_key", "test_secret_key")

    assert "AWS_ACCESS_KEY_ID" not in mock_environ
    assert "AWS_SECRET_ACCESS_KEY" not in mock_environ

    mock_chat_bedrock.assert_called_once()


@pytest.mark.asyncio
async def test_get_available_models(app, client):
    # Mock Config.shared()
    mock_config = MagicMock()
    mock_config.get_value.return_value = "mock_key"

    # Mock provider_warnings
    mock_provider_warnings = {
        ModelProviderName.openai: MagicMock(required_config_keys=["key1"]),
        ModelProviderName.amazon_bedrock: MagicMock(required_config_keys=["key2"]),
    }

    # Mock built_in_models
    mock_built_in_models = [
        KilnModel(
            name="model1",
            friendly_name="Model 1",
            family="",
            providers=[KilnModelProvider(name=ModelProviderName.openai)],
        ),
        KilnModel(
            name="model2",
            friendly_name="Model 2",
            family="",
            providers=[
                KilnModelProvider(name=ModelProviderName.amazon_bedrock),
                KilnModelProvider(name=ModelProviderName.ollama),
            ],
        ),
    ]

    # Mock connect_ollama
    mock_ollama_connection = OllamaConnection(
        message="Connected", models=["ollama_model1", "ollama_model2"]
    )

    with patch(
        "libs.server.kiln_server.provider_api.Config.shared", return_value=mock_config
    ), patch(
        "libs.server.kiln_server.provider_api.provider_warnings", mock_provider_warnings
    ), patch(
        "libs.server.kiln_server.provider_api.built_in_models", mock_built_in_models
    ), patch(
        "libs.server.kiln_server.provider_api.connect_ollama",
        return_value=mock_ollama_connection,
    ):
        response = client.get("/api/available_models")

    assert response.status_code == 200
    r = response.json()
    assert response.json() == [
        {
            "provider_id": "ollama",
            "provider_name": "Ollama",
            "models": [
                {"id": "model2", "name": "Model 2"},
            ],
        },
        {
            "provider_id": "openai",
            "provider_name": "OpenAI",
            "models": [{"id": "model1", "name": "Model 1"}],
        },
        {
            "provider_id": "amazon_bedrock",
            "provider_name": "Amazon Bedrock",
            "models": [{"id": "model2", "name": "Model 2"}],
        },
    ]


@pytest.mark.asyncio
async def test_get_available_models_ollama_exception(app, client):
    # Mock Config.shared()
    mock_config = MagicMock()
    mock_config.get_value.return_value = "mock_key"

    # Mock provider_warnings
    mock_provider_warnings = {
        ModelProviderName.openai: MagicMock(required_config_keys=["key1"]),
    }

    # Mock built_in_models
    mock_built_in_models = [
        KilnModel(
            name="model1",
            family="",
            friendly_name="Model 1",
            providers=[KilnModelProvider(name=ModelProviderName.openai)],
        ),
    ]

    # Mock connect_ollama to raise an HTTPException
    with patch(
        "libs.server.kiln_server.provider_api.Config.shared", return_value=mock_config
    ), patch(
        "libs.server.kiln_server.provider_api.provider_warnings", mock_provider_warnings
    ), patch(
        "libs.server.kiln_server.provider_api.built_in_models", mock_built_in_models
    ), patch(
        "libs.server.kiln_server.provider_api.connect_ollama",
        side_effect=HTTPException(status_code=500),
    ):
        response = client.get("/api/available_models")

    assert response.status_code == 200
    assert response.json() == [
        {
            "provider_id": "openai",
            "provider_name": "OpenAI",
            "models": [{"id": "model1", "name": "Model 1"}],
        },
    ]


def test_get_providers_models(client):
    response = client.get("/api/providers/models")
    assert response.status_code == 200

    data = response.json()
    assert "models" in data

    # Check if all built-in models are present in the response
    for model in built_in_models:
        assert model.name in data["models"]
        assert data["models"][model.name]["id"] == model.name
        assert data["models"][model.name]["name"] == model.friendly_name

    # Check if the number of models in the response matches the number of built-in models
    assert len(data["models"]) == len(built_in_models)

    if ModelName.llama_3_1_8b in data["models"]:
        assert data["models"][ModelName.llama_3_1_8b]["id"] == ModelName.llama_3_1_8b
        assert data["models"][ModelName.llama_3_1_8b]["name"] == "Llama 3.1 8B"
